require('module-alias/register')

const app = require('jsdoc/app');
const env = require('jsdoc/env');
const logger = require('jsdoc/util/logger');
const stripJsonComments = require('strip-json-comments');
const Promise = require('bluebird');
const fs = require('fs');
const path = require('path');

class JsDoc {
  constructor() {
    this.settings = {
      docs: [],
      packageJson: null,
      shouldExitWithError: false,
      tmpdir: null
    }

    let jsdocPath = __dirname

    if (fs.statSync(jsdocPath).isSymbolicLink()) {
      jsdocPath = path.resolve(path.dirname(jsdocPath), fs.readlinkSync(jsdocPath));
    }

    env.dirname = jsdocPath;
    env.pwd = process.cwd();
  }

  setVersionInfo() {
    env.version = {
      number: '',
      revision: ''
    }

    return this
  }

  getVersion() {
    return `JSDoc ${env.version.number} (${env.version.revision})`;
  }

  printVersion() {
    console.log(cli.getVersion());

    return Promise.resolve(0);
  }

  loadConfig(config) {
    const _ = require('underscore');
    const Config = require('jsdoc/config');

    const defaultOpts = {
      destination: './out/',
      encoding: 'utf8'
    };

    env.conf = new Config(config).get();
    env.opts = _.defaults(env.opts, env.conf.opts, defaultOpts);

    return this;
  }

  configureLogger() {
    const settings = this.settings
    const this$1 = this

    function recoverableError() {
      settings.shouldExitWithError = true;
    }

    function fatalError() {
      this$1.exit(1);
    }

    if (env.opts.debug) {
      logger.setLevel(logger.LEVELS.DEBUG);
    }
    else if (env.opts.verbose) {
      logger.setLevel(logger.LEVELS.INFO);
    }

    if (env.opts.pedantic) {
      logger.once('logger:warn', recoverableError);
      logger.once('logger:error', fatalError);
    }
    else {
      logger.once('logger:error', recoverableError);
    }

    logger.once('logger:fatal', fatalError);

    return this;
  }

  logStart() {
    logger.debug(this.getVersion());

    logger.debug('Environment info: %j', {
      env: {
        conf: env.conf,
        opts: env.opts
      }
    });

    return this
  }

  logFinish() {
    let delta;
    let deltaSeconds;

    if (env.run.finish && env.run.start) {
      delta = env.run.finish.getTime() - env.run.start.getTime();
    }

    if (delta !== undefined) {
      deltaSeconds = (delta / 1000).toFixed(2);
      logger.info('Finished running in %s seconds.', deltaSeconds);
    }
  }

  main() {
    this.scanFiles();

    if (env.sourceFiles.length === 0) {
      console.log('There are no input files to process.');

      return Promise.resolve(0);
    } else {
      return this.createParser()
        .parseFiles()
        .processParseResults()
        .then(() => {
          env.run.finish = new Date();

          return 0;
        });
    }
  }

  scanFiles() {
    const Filter = require('jsdoc/src/filter').Filter;

    let filter;

    env.opts._ = buildSourceList(this);

    // are there any files to scan and parse?
    if (env.conf.source && env.opts._.length) {
      filter = new Filter(env.conf.source);

      env.sourceFiles = app.jsdoc.scanner.scan(env.opts._,
        (env.opts.recurse ? env.conf.recurseDepth : undefined), filter);
    }

    return this;
  }

  createParser() {
    const handlers = require('jsdoc/src/handlers');
    const parser = require('jsdoc/src/parser');
    const plugins = require('jsdoc/plugins');

    app.jsdoc.parser = parser.createParser(env.conf.parser);

    if (env.conf.plugins) {
      env.conf.plugins = resolvePluginPaths(env.conf.plugins);
      plugins.installPlugins(env.conf.plugins, app.jsdoc.parser);
    }

    handlers.attachTo(app.jsdoc.parser);

    return this;
  }

  parseFiles() {
    const augment = require('jsdoc/augment');
    const borrow = require('jsdoc/borrow');
    const Package = require('jsdoc/package').Package;

    const settings = this.settings

    let docs;
    let packageDocs;

    settings.docs = docs = app.jsdoc.parser.parse(env.sourceFiles, env.opts.encoding);

    // If there is no package.json, just create an empty package
    packageDocs = new Package(settings.packageJson);
    packageDocs.files = env.sourceFiles || [];
    docs.push(packageDocs);

    logger.debug('Adding inherited symbols, mixins, and interface implementations...');
    augment.augmentAll(docs);
    logger.debug('Adding borrowed doclets...');
    borrow.resolveBorrows(docs);
    logger.debug('Post-processing complete.');

    app.jsdoc.parser.fireProcessingComplete(docs);

    return this;
  }

  processParseResults() {
    if (env.opts.explain) {
      this.dumpParseResults();

      return Promise.resolve();
    }
    else {
      this.resolveTutorials();

      return this.generateDocs();
    }
  }

  dumpParseResults() {
    console.log(dumper.dump(this.settings.docs));

    return this;
  }

  resolveTutorials() {
    const resolver = require('jsdoc/tutorial/resolver');

    if (env.opts.tutorials) {
      resolver.load(env.opts.tutorials);
      resolver.resolve();
    }

    return this;
  }

  generateDocs() {
    const path = require('jsdoc/path');
    const resolver = require('jsdoc/tutorial/resolver');
    const taffy = require('taffydb').taffy;

    let template;

    env.opts.template = (() => {
      const publish = env.opts.template || 'templates/default';
      const templatePath = path.getResourcePath(publish);

      // if we didn't find the template, keep the user-specified value so the error message is
      // useful
      return templatePath || env.opts.template;
    })();

    try {
      template = require(`${env.opts.template}/publish`);
    }
    catch (e) {
      logger.fatal(`Unable to load template: ${e.message}` || e);
    }

    // templates should include a publish.js file that exports a "publish" function
    if (template.publish && typeof template.publish === 'function') {
      let publishPromise;

      logger.info('Generating output files...');
      publishPromise = template.publish(
        taffy(this.settings.docs),
        env.opts,
        resolver.root
      );

      return Promise.resolve(publishPromise);
    }
    else {
      logger.fatal(`${env.opts.template} does not export a "publish" function. Global "publish" functions are no longer supported.`);
    }

    return Promise.resolve();
  }

  exit(exitCode, message) {
    if (exitCode > 0 && message) {
      console.error(message);
    }
    process.on('exit', () => { process.exit(exitCode); });
  }
}

function resolvePluginPaths(paths) {
  const path = require('jsdoc/path');
  const pluginPaths = [];

  // return a plugin path is resolved.
  function resolvePluginItem (plugin) {
    let pluginPath
    if (!path.isAbsolute(plugin)) {
      const basename = path.basename(plugin);
      const dirname = path.dirname(plugin);
      pluginPath = path.getResourcePath(dirname, basename);

      if (!pluginPath) {
        logger.error('Unable to find the plugin "%s"', plugin);
  
        return;
      }
    } else {
      pluginPath = plugin
    }
    return pluginPath
  }

  paths.forEach((plugin) => {
    /**
     * <plugin> => path:string or [path:string, options: {}]
     */
    if (Array.isArray(plugin)) {
      plugin[0] = resolvePluginItem(plugin[0])
    } else {
      plugin = resolvePluginItem(plugin)
    }

    pluginPaths.push(plugin);
  });

  return pluginPaths;
}

function readPackageJson(filepath) {
  const fs = require('jsdoc/fs');

  try {
    return stripJsonComments(fs.readFileSync(filepath, 'utf8'));
  }
  catch (e) {
    logger.error('Unable to read the package file "%s"', filepath);

    return null;
  }
}

function buildSourceList(jsdoc) {
  const Readme = require('jsdoc/readme');

  let packageJson;
  let readmeHtml;
  let sourceFile;
  let sourceFiles = env.opts._ ? env.opts._.slice(0) : [];

  if (env.conf.source && env.conf.source.include) {
    sourceFiles = sourceFiles.concat(env.conf.source.include);
  }

  // load the user-specified package/README files, if any
  if (env.opts.package) {
    packageJson = readPackageJson(env.opts.package);
  }
  if (env.opts.readme) {
    readmeHtml = new Readme(env.opts.readme).html;
  }

  // source files named `package.json` or `README.md` get special treatment, unless the user
  // explicitly specified a package and/or README file
  for (let i = 0, l = sourceFiles.length; i < l; i++) {
    sourceFile = sourceFiles[i];

    if (!env.opts.package && /\bpackage\.json$/i.test(sourceFile)) {
      packageJson = readPackageJson(sourceFile);
      sourceFiles.splice(i--, 1);
    }

    if (!env.opts.readme && /(\bREADME|\.md)$/i.test(sourceFile)) {
      readmeHtml = new Readme(sourceFile).html;
      sourceFiles.splice(i--, 1);
    }
  }

  jsdoc.settings.packageJson = packageJson;
  env.opts.readme = readmeHtml;

  return sourceFiles;
}

function excuteJsDoc(jsDocConfig, callback) {
  const jsDoc = new JsDoc(jsDocConfig)

  jsDoc.setVersionInfo()
    .loadConfig(jsDocConfig)
    .configureLogger()
    .logStart()
    .main()
    .then(errorCode => {
      if (!errorCode && jsDoc.settings.shouldExitWithError) {
        errorCode = 1;
      }

      if (errorCode === 0) {
        jsDoc.logFinish()
      }

      callback && callback(errorCode);
    });
}

module.exports = excuteJsDoc
